# -*- coding: UTF-8 -*-

"""StyleGAN architectures.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

from .base import StyleGAN
from _int import FMAP_SAMPLES, RES_INIT
from utils.latent_utils import gen_rand_latent_vars
from utils.custom_layers import Lambda, get_blur_op, NormalizeLayer, \
                                Conv2dEx, LinearEx, Conv2dBias

import copy

import numpy as np
import torch
from torch import nn
import torch.nn.functional as F

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

FMAP_G_INIT_FCTR = 1

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

class StyleMappingNetwork( nn.Module ):
  """Mapping Network for StyleGAN architecture."""
  def __init__( self,
                len_latent = 512,
                len_dlatent = 512,
                num_fcs = 8,
                lrmul = .01,
                nl = nn.LeakyReLU( negative_slope = .2 ),
                equalized_lr = True,
                normalize_z = True ):
    super( StyleMappingNetwork, self ).__init__()

    if normalize_z:
      self.preprocess_z = nn.Sequential(
        Lambda( lambda x: x.view( -1, len_latent ) ),
        NormalizeLayer( 'PixelNorm' )
      )
    else:
      self.preprocess_z = Lambda( lambda x: x.view( -1, len_latent ) )

    self.dims = np.linspace( len_latent, len_dlatent, num_fcs + 1 ).astype( np.int64 )
    self.fc_mapping_model = nn.Sequential( )
    for seq_n in range( num_fcs ):
      self.fc_mapping_model.add_module(
        'fc_' + str( seq_n ),
        LinearEx( nin_feat = self.dims[seq_n], nout_feat = self.dims[seq_n+1],
                  init = 'He', init_type = 'StyleGAN', gain_sq_base = 2.,
                  equalized_lr = equalized_lr, lrmul = lrmul )
      )
      self.fc_mapping_model.add_module( 'nl_' + str( seq_n ), nl )

  def forward( self, x ):
    return self.fc_mapping_model( self.preprocess_z( x ) )


class StyleConditionedMappingNetwork( nn.Module ):
  """Class-conditioned version of Mapping Network for StyleGAN architecture."""
  def __init__( self,
                num_classes,
                len_latent = 512,
                len_dlatent = 512,
                num_fcs = 8,
                lrmul = .01,
                nl = nn.LeakyReLU( negative_slope = .2 ),
                equalized_lr = True,
                normalize_z = True,
                embed_cond_vars = True ):
    super( StyleConditionedMappingNetwork, self ).__init__()

    self.len_latent = len_latent
    self.num_classes = num_classes

    self.embed_cond_vars = embed_cond_vars
    if embed_cond_vars:
      self.class_embedding = LinearEx( nin_feat = num_classes, nout_feat = len_latent,
                                       init = None, init_type = 'Standard Normal', include_bias = False )

    self.dims = np.linspace( len_latent, len_dlatent, num_fcs ).astype( np.int64 )
    self.dims = np.insert( self.dims, 0, 2*len_latent if self.embed_cond_vars else len_latent + num_classes )
    self.fc_mapping_model = nn.Sequential( )
    if normalize_z:
      self.fc_mapping_model.add_module( 'pixelnorm', NormalizeLayer( 'PixelNorm' ) )
    for seq_n in range( num_fcs ):
      self.fc_mapping_model.add_module(
        'fc_' + str( seq_n ),
        LinearEx( nin_feat = self.dims[seq_n], nout_feat = self.dims[seq_n+1],
                  init = 'He', init_type = 'StyleGAN', gain_sq_base = 2.,
                  equalized_lr = equalized_lr, lrmul = lrmul )
      )
      self.fc_mapping_model.add_module( 'nl_' + str( seq_n ), nl )

  def forward( self, x, y ):
    y = y.view( -1, self.num_classes )
    if self.embed_cond_vars:
      y = self.class_embedding( y )
    return self.fc_mapping_model( torch.cat( ( x.view( -1, self.len_latent ), y, ), dim = 1 ) )


class StyleAddNoise( nn.Module ):
  """Simple `nn.Module` that adds weighted uncorrelated Gaussian noise to a layer of feature maps."""
  def __init__( self, nf ):
    super( StyleAddNoise, self ).__init__()

    self.noise_weight = nn.Parameter( torch.FloatTensor( 1, nf, 1, 1 ).fill_( 0 ) )

  def forward( self, x, noise = None ):
    if self.training or noise is None:
      # for training mode or when user does not supply noise in evaluation mode:
      return x + self.noise_weight * \
        torch.randn( x.shape[0], 1, x.shape[2], x.shape[3], dtype = torch.float32, device = x.device )
    else:
      # for if when user supplies noise in evaluation mode:
      return x + self.noise_weight * noise

# ............................................................................ #
# Generator:
# ----------

# TODO: Implement more efficient "recursive structure" by Tero Karras for training (see their GitHub implementation)
class StyleGenerator( StyleGAN ):
  """StyleGAN (Karras et al. 2019) Generator

  Emulates most recent official implementation.
  """
  def __init__( self,
                final_res,
                latent_distribution = 'normal',
                len_latent = 512,
                len_dlatent = 512,
                mapping_num_fcs = 8,
                mapping_lrmul = .01,
                use_instancenorm = True,
                use_noise = True,
                upsampler = nn.Upsample( scale_factor = 2, mode = 'nearest' ),
                blur_type = None,
                nl = nn.LeakyReLU( negative_slope = .2 ),
                num_classes = 0,
                equalized_lr = True,
                normalize_z = True,
                use_pixelnorm = False,
                pct_mixing_reg = .9,
                truncation_trick_params = { 'beta': .995, 'psi': .7, 'cutoff_stage': 4 } ):

    super( self.__class__, self ).__init__( final_res )

    self.gen_layers = nn.ModuleList( )

    self.upsampler = upsampler
    self.upsampler_skip_connection = \
      lambda xb: F.interpolate( xb, scale_factor = 2, mode = 'nearest' )  # keep fading-in layers simple

    self.gen_blur_type = blur_type

    self.nl = nl

    self.equalized_lr = equalized_lr

    self.pct_mixing_reg = pct_mixing_reg
    self._use_mixing_reg = True if pct_mixing_reg else False

    self.latent_distribution = latent_distribution
    self.len_latent = len_latent
    self.len_dlatent = len_dlatent
    assert isinstance( num_classes, int )
    self.num_classes = num_classes

    # Mapping Network initialization:
    if not num_classes:
      self.z_to_w = StyleMappingNetwork(
        len_latent = len_latent,
        len_dlatent = len_dlatent,
        num_fcs = mapping_num_fcs,
        lrmul = mapping_lrmul,
        nl = nl,
        equalized_lr = equalized_lr,
        normalize_z = normalize_z
      )
    else:
      self.z_to_w = StyleConditionedMappingNetwork(
        num_classes,
        len_latent = len_latent,
        len_dlatent = len_dlatent,
        num_fcs = mapping_num_fcs,
        lrmul = mapping_lrmul,
        nl = nl,
        equalized_lr = equalized_lr,
        normalize_z = normalize_z
      )

    _fmap_init = len_latent * FMAP_G_INIT_FCTR

    # initializing the input to 1 has about the same effect as applyng PixelNorm to the input
    self.const_input = nn.Parameter(
      torch.FloatTensor( 1, _fmap_init, RES_INIT, RES_INIT ).fill_( 1 )
    )

    self._use_noise = use_noise
    self._trained_with_noise = use_noise
    if use_noise:
      conv = Conv2dEx( ni = _fmap_init, nf = self.fmap, ks = 3,
                       stride = 1, padding = 1, init = 'He', init_type = 'StyleGAN',
                       gain_sq_base = 2., equalized_lr = equalized_lr, include_bias = False )
      noise = [
        StyleAddNoise( nf = _fmap_init ),
        StyleAddNoise( nf = self.fmap ),
      ]
      bias = (
        [ Conv2dBias( nf = _fmap_init ) ],
        [ Conv2dBias( nf = self.fmap ) ],
      )
    else:
      conv = Conv2dEx( ni = _fmap_init, nf = self.fmap, ks = 3,
                       stride = 1, padding = 1, init = 'He', init_type = 'StyleGAN',
                       gain_sq_base = 2., equalized_lr = equalized_lr, include_bias = True )
      # noise = ( [], [], )
      noise = [ None, None ]
      bias = ( [], [], )  # NOTE: without noise, the bias would get directly added to the constant input, so the constant input can just learn this bias,
                          #       so theoretically, there shouldn't be a need to include the bias either. There may be numerical approximation problems from backprop, however.

    norms = []
    self.use_pixelnorm = use_pixelnorm
    if use_pixelnorm:
      norms.append( NormalizeLayer( 'PixelNorm' ) )
    self.use_instancenorm = use_instancenorm
    if use_instancenorm:
      norms.append( NormalizeLayer( 'InstanceNorm' ) )

    w_to_styles = (
      LinearEx( nin_feat = self.z_to_w.dims[ -1 ], nout_feat = 2 * _fmap_init,
                init = 'He', init_type = 'StyleGAN', gain_sq_base = 1., equalized_lr = equalized_lr ),
      LinearEx( nin_feat = self.z_to_w.dims[ -1 ], nout_feat = 2 * self.fmap,
                init = 'He', init_type = 'StyleGAN', gain_sq_base = 1., equalized_lr = equalized_lr ),
    )
    assert 0. <= truncation_trick_params[ 'beta' ] <= 1.
    self.w_ewma_beta = truncation_trick_params[ 'beta' ]
    self._w_eval_psi = truncation_trick_params[ 'psi' ]  # allow psi to be any number you want, perhaps worthy of experimentation
    assert ( ( isinstance( truncation_trick_params[ 'cutoff_stage' ], int ) and \
                0 < truncation_trick_params[ 'cutoff_stage' ] <= int( np.log2( self.final_res ) ) - 2 ) or \
                truncation_trick_params[ 'cutoff_stage' ] is None )
    self._trunc_cutoff_stage = truncation_trick_params[ 'cutoff_stage' ]
    # set the below to `False` if you want to turn off during evaluation mode
    self.use_truncation_trick = True if self._trunc_cutoff_stage else False
    self.w_ewma = None

    self.gen_layers.append(
      nn.ModuleList( [
        None,
        noise[0],
        nn.Sequential( *bias[0], nl, *norms ),
        w_to_styles[0]
      ] )
    )
    self.gen_layers.append(
      nn.ModuleList( [
        conv,
        noise[1],
        nn.Sequential( *bias[1], nl, *norms ),
        w_to_styles[1]
      ] )
    )

    self.prev_torgb = None
    self._update_torgb( ni = self.fmap )

  def increase_scale( self ):
    """Use this to increase scale during training or for initial resolution."""
    # update metadata
    if not self.scale_inc_metadata_updated:
      super( self.__class__, self ).increase_scale()
    else:
      self.scale_inc_metadata_updated = False

    blur_op = get_blur_op( blur_type = self.gen_blur_type, num_channels = self.fmap ) if \
              self.gen_blur_type is not None else None

    self.gen_layers.append(
      self.get_conv_layer( ni = self.fmap_prev, upsample = True, blur_op = blur_op )
    )
    self.gen_layers.append(
      self.get_conv_layer( ni = self.fmap )
    )

    self.prev_torgb = copy.deepcopy( self.torgb )
    self._update_torgb( ni = self.fmap )

  def get_conv_layer( self, ni, upsample = False, blur_op = None, append_nl = True ):
    upsampler = []
    if upsample:
      upsampler.append( self.upsampler )

    if self.use_noise or blur_op is not None:
      conv = Conv2dEx( ni = ni, nf = self.fmap, ks = 3, stride = 1, padding = 1,
                       init = 'He', init_type = 'StyleGAN', gain_sq_base = 2.,
                       equalized_lr = self.equalized_lr, include_bias = False )
      bias = [ Conv2dBias( nf = self.fmap ) ]
    else:
      conv = Conv2dEx( ni = ni, nf = self.fmap, ks = 3, stride = 1, padding = 1,
                       init = 'He', init_type = 'StyleGAN', gain_sq_base = 2.,
                       equalized_lr = self.equalized_lr, include_bias = True )
      bias = []

    blur = []
    if blur_op is not None:
      assert isinstance( blur_op, nn.Module )
      blur.append( blur_op )

    noise = None
    if self.use_noise:
      noise = StyleAddNoise( nf = self.fmap )

    nl = []
    if append_nl:
      nl.append( self.nl )

    norms = []
    if self.use_pixelnorm:
      norms.append( NormalizeLayer( 'PixelNorm' ) )
    if self.use_instancenorm:
      norms.append( NormalizeLayer( 'InstanceNorm' ) )

    w_to_style = LinearEx( nin_feat = self.z_to_w.dims[ -1 ], nout_feat = 2*self.fmap,
                           init = 'He', init_type = 'StyleGAN', gain_sq_base = 1.,
                           equalized_lr = self.equalized_lr )

    return nn.ModuleList( [ nn.Sequential( *upsampler, conv, *blur ),
                            noise,
                            nn.Sequential( *( bias + nl + norms ) ),
                            w_to_style ] )

    # return nn.ModuleList( [ nn.Sequential( *upsampler, conv, *( blur + noise + bias + nl + norms ) ), w_to_style ] )

  def _update_torgb( self, ni ):
    self.torgb = Conv2dEx( ni = ni, nf = FMAP_SAMPLES, ks = 1, stride = 1,
                           padding = 0, init = 'He', init_type = 'StyleGAN',
                           gain_sq_base = 1., equalized_lr = self.equalized_lr )

  def train( self, mode = True ):
    """Overwritten to turn on mixing regularization (if > 0%) during training mode."""
    super( self.__class__, self ).train( mode = mode )

    self._use_noise = self._trained_with_noise
    self._use_mixing_reg = True if self.pct_mixing_reg else False

  def eval( self ):
    """Overwritten to turn off mixing regularization during evaluation mode."""
    super( self.__class__, self ).eval( )

    self._use_mixing_reg = False

  def to( self, *args, **kwargs ):
    """Overwritten to allow for non-Parameter objects' Tensors to be sent to the appropriate device."""
    super( self.__class__, self ).to( *args, **kwargs )

    for arg in args:
      if arg in ( 'cpu', 'cuda', ) or isinstance( arg, torch.device ):
        if self.w_ewma is not None:
          self.w_ewma = self.w_ewma.to( arg )
          break

  @property
  def use_noise( self ):
    return self._use_noise

  @use_noise.setter
  def use_noise( self, mode ):
    """Allows for optionally evaluating without noise inputs."""
    if self.training:
      raise Exception( 'Once use_noise argument is set, it cannot be changed' + \
                       ' for training purposes. It can, however, be changed in eval mode.' )
    elif not self._trained_with_noise:
      raise Exception( 'Model was not trained with noise, so cannot use noise in eval mode.' )
    else:
      self._use_noise = mode

  @property
  def w_eval_psi( self ):
    return self._w_eval_psi

  @w_eval_psi.setter
  def w_eval_psi( self, new_w_eval_psi ):
    """Change this to your choosing (but only in evaluation mode), optionally allowing for |psi| to be > 1."""
    if not self.training:
      self._w_eval_psi = new_w_eval_psi
    else:
      raise Exception( 'Can only alter psi value for truncation trick on w during evaluation mode.' )

  @property
  def trunc_cutoff_stage( self ):
    return self._trunc_cutoff_stage

  @trunc_cutoff_stage.setter
  def trunc_cutoff_stage( self, new_trunc_cutoff_stage ):
    """Change this to your choosing (but only in evaluation mode)."""
    if not self.training:
      _final_stage = int( np.log2( self.final_res ) ) - 1
      if ( isinstance( new_trunc_cutoff_stage, int ) and \
        0 < new_trunc_cutoff_stage <= _final_stage ) or new_trunc_cutoff_stage is None:
        self._trunc_cutoff_stage = new_trunc_cutoff_stage
      else:
        message = f'Input cutoff stage for truncation trick on w must be of type `int` in range (0,{_final_stage}] or `None`.'
        raise ValueError( message )
    else:
      raise Exception( 'Can only alter cutoff stage for truncation trick on w during evaluation mode.' )

  def forward( self, x, x_mixing = None, style_mixing_stage:int = None, noise = None ):
    # TODO: Implement the ability to style-mix more than just 2 styles in eval mode
    # TODO: Implement the ability to input the disentangled latent variable w directly

    cutoff_idx = None
    # Training Mode Only:
    if self._use_mixing_reg:
      if np.random.rand() < self.pct_mixing_reg:
        if self.alpha != 0:
          cutoff_idx = torch.randint( 1, 2*self.scale_stage, ( 1, ) ).item()
        else:
          cutoff_idx = torch.randint( 1, 2*self.scale_stage - 2, ( 1, ) ).item()

    x = self.z_to_w( x )
    bs = x.shape[0]

    if self.use_truncation_trick:
      # Training Mode Only:
      if self.training:
        if self.w_ewma is None:
          self.w_ewma = x.detach().clone().mean( dim = 0 )
        else:
          with torch.no_grad():
            # TODO: Implement a memory-efficient method to compute this for the ewma generator
            #       (currently just using the same average w for the generator and the ewma generator)
            self.w_ewma = x.mean( dim = 0 ) * ( 1. - self.w_ewma_beta ) + \
                          self.w_ewma * ( self.w_ewma_beta )
      # Evaluation Mode Only:
      elif self.trunc_cutoff_stage is not None:
        x = self.w_ewma.expand_as( x ) + self.w_eval_psi * ( x - self.w_ewma.expand_as( x ) )

    out = self.const_input.expand( bs, -1, -1, -1 )

    if self.fade_in_phase:
      for n, layer in enumerate( self.gen_layers[ :-2 ] ):
        if n:
          out = layer[ 0 ]( out )
        if self.use_noise:
          out = layer[ 1 ]( out, noise = noise[ n ] if noise is not None else None )
        out = layer[ 2 ]( out )

        if n == cutoff_idx:
          # TODO: Implement embedding-style conditioning from "Which Training Methods for
          #       GANs do actually Converge" & discriminator conditioning.
          x = gen_rand_latent_vars( num_samples = bs, length = self.len_latent,
                                    distribution = self.latent_distribution, device = x.device )
          x.requires_grad_( True )
          x = self.z_to_w( x )

        y = layer[ 3 ]( x ).view( -1, 2, layer[ 3 ].nout_feat // 2, 1, 1 )
        out = out * ( y[ :, 0 ].contiguous().add( 1 ) ) + \
              y[ :, 1 ].contiguous()  # add 1 for skip-connection effect

      # TODO: there should be a cleaner way to do the fading-in part while remaining memory-efficient...
      n += 1
      if n == cutoff_idx:
        x = gen_rand_latent_vars( num_samples = bs, length = self.len_latent,
                                  distribution = self.latent_distribution, device = x.device )
        x.requires_grad_( True )
        x = self.z_to_w( x )
      y = self.gen_layers[ -2 ][ 3 ]( x ).view( -1, 2, self.gen_layers[ -2 ][ 3 ].nout_feat // 2, 1, 1 )

      n += 1
      if n == cutoff_idx:
        x = gen_rand_latent_vars( num_samples = bs, length = self.len_latent,
                                  distribution = self.latent_distribution, device = x.device )
        x.requires_grad_( True )
        x = self.z_to_w( x )
      yf = self.gen_layers[ -1 ][ 3 ]( x ).view( -1, 2, self.gen_layers[ -1 ][ 3 ].nout_feat // 2, 1, 1 )

      if self.use_noise:
        return self.upsampler_skip_connection( self.prev_torgb( out ) ) * ( 1. - self.alpha ) + \
          self.torgb(
            self.gen_layers[ -1 ][ 2 ]( self.gen_layers[ -1 ][ 1 ]( self.gen_layers[ -1 ][ 0 ](
              self.gen_layers[ -2 ][ 2 ]( self.gen_layers[ -2 ][ 1 ]( self.gen_layers[ -2 ][ 0 ]( out ), noise = noise[ -2 ] if noise is not None else None ) ) * ( y[ :, 0 ].contiguous().add( 1 ) ) + y[ :, 1 ].contiguous()
            ), noise = noise[ -1 ] if noise is not None else None ) ) * ( yf[ :, 0 ].contiguous().add( 1 ) ) + yf[ :, 1 ].contiguous()
          ) * ( self.alpha )
      else:
        return self.upsampler_skip_connection( self.prev_torgb( out ) ) * ( 1. - self.alpha ) + \
          self.torgb(
            self.gen_layers[ -1 ][ 2 ]( self.gen_layers[ -1 ][ 0 ](
              self.gen_layers[ -2 ][ 2 ]( self.gen_layers[ -2 ][ 0 ]( out ) ) * ( y[ :, 0 ].contiguous().add( 1 ) ) + y[ :, 1 ].contiguous()
            ) ) * ( yf[ :, 0 ].contiguous().add( 1 ) ) + yf[ :, 1 ].contiguous()
          ) * ( self.alpha )

    else:
      for n, layer in enumerate( self.gen_layers ):
        if n:
          out = layer[ 0 ]( out )
        if self.use_noise:
          out = layer[ 1 ]( out, noise = noise[ n ] if noise is not None else None )
        out = layer[ 2 ]( out )

        # Training Mode Only:
        if n == cutoff_idx:
          # TODO: Implement embedding-style conditioning from "Which Training Methods for
          #       GANs do actually Converge" & discriminator conditioning.
          x = gen_rand_latent_vars( num_samples = bs, length = self.len_latent,
                                    distribution = self.latent_distribution, device = x.device )
          x.requires_grad_( True )
          x = self.z_to_w( x )

        # Evaluation Mode Only:
        if n == style_mixing_stage:
          assert ( style_mixing_stage and not self.training and isinstance( x_mixing, torch.Tensor ) )
          x = self.z_to_w( x_mixing )
          # the new z that is sampled for style-mixing is already de-truncated
          if self.use_truncation_trick and self.trunc_cutoff_stage is not None and n < 2*self.trunc_cutoff_stage:
            x = self.w_ewma.expand_as( x ) + self.w_eval_psi * ( x - self.w_ewma.expand_as( x ) )
        elif self.use_truncation_trick and not self.training and self.trunc_cutoff_stage is not None and n == 2*self.trunc_cutoff_stage:
          # de-truncate w for higher resolutions; more memory-efficient than defining 2 w's
          x = ( x - self.w_ewma.expand_as( x ) ).div( self.w_eval_psi ) + self.w_ewma.expand_as( x )

        y = layer[ 3 ]( x ).view( -1, 2, layer[ 3 ].nout_feat // 2, 1, 1 )
        out = out * ( y[ :, 0 ].contiguous().add( 1 ) ) + \
              y[ :, 1 ].contiguous()  # add 1 for skip-connection effect

      return self.torgb( out )


# -*- coding: UTF-8 -*-

"""Latent Space-specific utilities.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

import torch

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

def gen_rand_latent_vars( num_samples, length, distribution = 'normal',
                          device = 'cuda' if torch.cuda.is_available() else 'cpu' ):
  if distribution == 'normal':
    z = torch.randn( num_samples, length, dtype = torch.float32, device = device )
  elif distribution == 'uniform':
    z = torch.rand( num_samples, length, dtype = torch.float32, device = device )

  return z

def concat_rand_classes_to_z( z, num_classes, z_labels = None, \
                              device = 'cuda' if torch.cuda.is_available() else 'cpu' ):
  assert z_labels.shape == ( len( z ), 1, )

  if z_labels is None:
    z_labels = torch.randint( 0, num_classes, ( len( z ), 1, ), dtype = torch.int64, device = device )

  labels_one_hot = torch.zeros( len( z ), num_classes, dtype = torch.float32, device = device )
  labels_one_hot.scatter_( 1, z_labels, 1 )

  return torch.cat( ( z, labels_one_hot ), dim = 1 )


# -*- coding: UTF-8 -*-

"""A Style-Based Generator Architecture for Generative Adversarial Networks.
   (Karras et al. 2019)

Default configuration exactly emulates the most recent official implementation.

That being said, there are plenty of hyperparameter options/settings that you
can change in config.py to alter your StyleGAN to your liking.
"""

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

import os; import sys
sys.path.append(
  os.path.abspath( os.path.join( os.path.dirname( __file__ ), '..' ) )
)

from . import base
from . import architectures
from . import learner

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#


# -*- coding: UTF-8 -*-

"""Configures package for the user-specified dataset. Contains useful utilities.

Run 'python data_config.py [dataset] [dataset_dir]' on the command-line to
configure (or re-configure) this package for the dataset you wish to use, and
optionally specify any set of '--[keyword argument]' to override any default
keyword arguments pertaining to image transforms or validation set creation.

  Typical usage examples:

  $ python data_config.py FFHQ path/to/datasets/ffhq  [OR]
  $ python data_config.py FFHQ path/to/datasets/ffhq --enable_mirror_augmentation  [OR]
  $ python data_config.py FFHQ path/to/datasets/ffhq --include_valid_set=False

After running this module (as well as running config.py), you will be able to
construct (at any later time) a set of torch Datasets and DataLoaders that is
specific to your input dataset configuration & hyperparameters by running
the "prepare_dataset(...)" and "prepare_dataloader(...)" functions respectively,
from the utils.data_utils.py module.

This module can also be imported externally and used for its various utilities.

If you would like to see a list of what each argument does,
run 'python data_config.py [dataset] [dataset_dir] -h' on the command-line.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

import sys
if '_int' in sys.modules or __name__ == '__main__':
  from _int import str2bool, get_current_configuration, \
                   TRAINING_SET_DIRNAME, VALID_SET_DIRNAME, \
                   TORCHVISION_CATERED_DATASETS
elif 'gan_lab._int' in sys.modules:
  from ._int import str2bool, get_current_configuration, \
                    TRAINING_SET_DIRNAME, VALID_SET_DIRNAME, \
                    TORCHVISION_CATERED_DATASETS

import os
import argparse
from typing import Union
from functools import partial
from pathlib import Path
import shutil
import pickle

import numpy as np
import torch
from torchvision.datasets.folder import *
from torchvision import transforms

from PIL import Image

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

# TODO: Implement ability to transform without having to use PIL transforms/convert to PIL Image object
USE_PIL_TRANSFORMS = True

DEFAULT_STATS = { 'mean' : [ 0.5, 0.5, 0.5 ], 'std' : [ 0.5, 0.5, 0.5 ] }
LSUN_BEDROOM_STATS = { 'mean' : [ 0.5, 0.5, 0.5 ], 'std' : [ 0.5, 0.5, 0.5 ] }
CIFAR10_STATS = { 'mean' : [ .491, .482, .447 ], 'std' : [ .247, .243, .261 ] }
CELEBAHQ_STATS = { 'mean' : [ .5, .5, .5 ], 'std' : [ .5, .5, .5 ] }

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

# ---------------------------------------------------------------------------- #

# Validation Set creation:
# ------------------------

def make_valid_set( dataset_dir:Path, training_set_dirname = 'train', valid_set_dirname = 'valid', valid_set_pct = None, valid_set_idxs:np.integer = None ):
  """Partitions dataset into training set and validation set based on either input % validation or validation indices of training set."""
  try:
    dataset_dir = Path( dataset_dir )
  except:
    raise TypeError( 'Input dataset path must be convertible to a "pathlib.Path" instance.' )

  assert dataset_dir.exists()

  if valid_set_pct is not None:
    assert isinstance( valid_set_pct, ( float, int, ) )
    assert ( 0 <= valid_set_pct <= 1 )
  _bool_valid_idx_np = False
  if valid_set_idxs is not None:
    assert isinstance( valid_set_idxs, np.integer )
    if valid_set_idxs.size > 0: _bool_valid_idx_np = True

  training_set_path = dataset_dir/training_set_dirname
  training_set_path_str = str( training_set_path )
  valid_set_path = dataset_dir/valid_set_dirname
  valid_set_path_str = str( valid_set_path )

  if valid_set_pct is not None or valid_set_idxs is not None:

    if valid_set_pct != _bool_valid_idx_np:  # xor

      if not ( training_set_path.exists() and training_set_path.is_dir() ):
        print( f'Creating training set at "{training_set_path_str}"...\n' )
        training_set_path.mkdir( exist_ok = False )
        for f in dataset_dir.iterdir():
          if f.is_file():
            shutil.move( str( f ), training_set_path_str )

      if valid_set_path.exists() and valid_set_path.is_dir():
        print( f'Validation set already exists in directory "{valid_set_path_str}".\n' + \
               f'Moving validation set data back to "{training_set_path_str}" before re-making validation set...\n' )
        for f in valid_set_path.iterdir():
          if f.is_file():
            shutil.move( str( f ), training_set_path_str )

      dataset_sz = len( [ f for f in training_set_path.iterdir() if f.is_file() ] )

      valid_set_path.mkdir( exist_ok = True )
      if valid_set_pct:
        valid_set_idxs = np.random.choice(
          dataset_sz, size = round( valid_set_pct * dataset_sz ), replace = False
        )
      print( f'{len(valid_set_idxs)} images will now be moved from "{training_set_path_str}" to "{valid_set_path_str}"...\n' )
      for n, f in enumerate( training_set_path.iterdir() ):
        if n in valid_set_idxs and f.is_file():
          shutil.move( str( f ), valid_set_path_str )

    elif valid_set_pct and _bool_valid_idx_np:
      raise ValueError( 'Cannot specify both a validation pct and array of indices. Please choose one.' )

  else:
    if not ( training_set_path.exists() and training_set_path.is_dir() ) or \
       not ( valid_set_path.exists() and valid_set_path.is_dir() ):
      raise ValueError( 'validation set desired but validation set cannot be created because' + \
                        ' no % validation or validation indices specified' )
    raise RuntimeWarning( 'No % validation or validation indices specified.' + \
                          ' Therefore nothing has been done with regards to validation set creation (or re-creation).\n' )

  imgs_dir = training_set_path

  return imgs_dir

# ---------------------------------------------------------------------------- #

# Data Normalization:
# -------------------

# NOTE: Input stats_type = 'default' into the function below to map pixel intensity
#       value 0 to -1 and pixel intensity value 255 to 1.
def normalize_center_standardize( dataset = 'default', normalize = False, standardize = False ):
  """Normalize, center, and/or standardize data to [-1,1].
     These assume that your stats (constants listed above) are for the normalized (i.e. range [0,1]) versions of these images.
  """
  dataset = dataset.casefold()

  if standardize:
    if dataset == 'default':
      stats = DEFAULT_STATS
    elif dataset in ( 'lsun bedrooms', 'lsun-bedrooms', 'lsunbedrooms', ):
      stats = LSUN_BEDROOM_STATS
    elif dataset in ( 'cifar-10', 'cifar10', 'cifar 10', ):
      stats = DEFAULT_STATS  # CIFAR10_STATS
    elif dataset in ( 'celeba-hq', 'celebahq', 'celeba hq', ):
      stats = CELEBAHQ_STATS
    # elif dataset == 'ffhq':
    #   raise NotImplementedError( 'FFHQ dataset configuration support not yet implemented.' )
    else:
      raise NotImplementedError(
        "input dataset's statistics not yet implemented. Please use dataset = `'default'`" + \
        " instead for default (and recommended) standardization, or equivalently just set standardization = `False`."
      )

  _nfctr = 1.
  if normalize:
    _nfctr = 255.

  if standardize:
    mean = [ m*_nfctr for m in stats[ 'mean' ] ]
    std = [ s*_nfctr for s in stats[ 'std' ] ]
  else:
    mean = [ m*_nfctr for m in DEFAULT_STATS[ 'mean' ] ]
    std = [ s*_nfctr for s in DEFAULT_STATS[ 'std' ] ]

  return mean, std

def normalize_notorch( dataset ):
  """Normalize data when torch is unavailable."""
  raise NotImplementedError( '"normalize_notorch( data )" not yet implemented.' )

# ---------------------------------------------------------------------------- #

# Misc. Utilities:
# ----------------

def get_dataset_img_extension( imgs_dir ):
  """Quick (non-comprehensive) function to get the extension of the dataset's data."""
  fs = Path( imgs_dir ).iterdir()
  f0 = next( fs )
  while not f0.is_file():
    f0 = next( fs )
  fs = iter( reversed( list( Path( imgs_dir ).iterdir() ) ) )
  ff = next( fs )
  while not ff.is_file():
    ff = next( fs )
  if f0.suffix.casefold() == ff.suffix.casefold():
    return f0.suffix.casefold()
  else:
    raise Exception( "Unable to identify the extension of the dataset's data" )

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

# Set up and configure your input dataset:
# ----------------------------------------

if __name__ == '__main__':

  parser = argparse.ArgumentParser( prog = 'GAN Lab', description = 'StyleGAN, ProGAN, and ResNet GANs (dataset configuration)' )

  if len( sys.argv ) < 3:
    raise TypeError( 'data_config.py requires atleast 2 arguments (the dataset name and dataset directory), and then optional keyword arguments (prefixed with "--").\n' + \
                     ' Example: "$ python data_config.py FFHQ path/to/datasets/ffhq"' )

  # .......................................................................... #

  # POSITIONAL (REQUIRED) ARGUMENTS:
  # --------------------------------

  parser.add_argument( 'dataset', type = str.casefold, help = 'input dataset name (e.g. "FFHQ") used to train the model' )
  parser.add_argument( 'dataset_dir', type = Path, help = 'root directory where training data is stored' )

  # .......................................................................... #

  # KEYWORD (OPTIONAL) ARGUMENTS:
  # -----------------------------

  # Image Data Transforms
  # transforms performed in the order listed below; they are not implemented if not specified:
  parser.add_argument( '--center_crop_dim', type = int, help = 'square size of image after center cropping' )
  parser.add_argument( '--random_crop_dim', type = int, help = 'square size of image after random cropping' )
  parser.add_argument( '--dataset_downsample_type', type = str.casefold, default = 'average', choices = [ 'nearest', 'average', 'box', 'bilinear' ], \
    help = 'type of downsampling to conduct on dataset to get it to the target resolution for the model; for ResNet GANs, this target resolution is config.res_samples;' + \
           ' for ProGAN and StyleGAN, this target resolution is every resolution stage from config.init_res to the final resolution (i.e. config.res_samples) inclusive;' + \
           ' (side-note: if config has not yet been initialized, the ds_transforms attribute of data_config will show Resize as a "functools.partial" function until config is initialized)' )
  parser.add_argument( '--enable_mirror_augmentation', type = str2bool, nargs = '?', const = True, default = False, \
    help = 'whether to enable mirror augmentation (i.e. horizontal-flip/flip-lr augmentation)' )
  parser.add_argument( '--standardize', type = str2bool, nargs = '?', const = True, default = False, \
    help = 'optional (not recommended) standardization of pixels based on dataset statistics;' + \
           ' otherwise, map lowest possible pixel value to -1. and highest possible pixel value to 1 by default.' )

  # Validation Set Creation:
  parser.add_argument( '--include_valid_set', type = str2bool, nargs = '?', const = True, default = True, \
    help = "whether the input dataset should include a validation set during training (and therefore create a validation DataLoader);" + \
           " this flag also signals to create (or re-create) a validation set based on the 2 arguments below;" + \
           " however, if the dataset already comes pre-installed with a validation set (like with many torchvision-catered" + \
           " datasets such as LSUN Bedrooms or CIFAR-10), then all aspects considering validation set creation (or re-creation) are ignored," + \
           " and the 2 arguments below mean nothing (but the flag can still be used to signal that you want to train with the pre-installed validation set)" )
  parser.add_argument( '--valid_set_pct', type = float, default = .01, \
    help = "percent of training set to convert into validation set; ignored if --include_valid_set (above) == `False`;" + \
           " else if validation set already exists in dataset_dir (top), then this replaces the original validation set with this" + \
           " new one and moves the original validation set data back into the training set (you do not want to do this if goal is reproducibility," +\
           " in which case, set this equal to `None` (and keep the below argument equal to `None` as well) and nothing will happen)" )
  parser.add_argument( '--valid_set_idxs', type = np.integer, default = None, \
    help = "set of indices in the training set that will be used to create validation set; ignored if --include_valid_set (2 above) == `False`;" + \
           " else if validation set already exists in dataset_dir (top), then this replaces the original validation set with this" + \
           " new one and moves the original validation set data back into the training set (you do not want to use this if goal is reproducibility);" + \
           " overwrites --valid_set_pct (above)" )

  parser.add_argument( '--ds_loader', type = str.casefold, \
    help = "name of any possible loader (a function) one wants to use to load a sample given its path (see utils.data_utils.py for built-in loader(s));" + \
           " only necessary if using the generic DatasetFolderSingleClass class for your torch Dataset (see utils.data_utils.py);" + \
           " typically, as long as the data's extension is a generic image extension or if using one of torchvision's curated dataset classes," + \
           " DatasetFolderSingleClass will not be instantiated and a loader will thus not be necessary" )

  # .......................................................................... #

  data_config = parser.parse_args( )

  config = get_current_configuration( 'config', raise_exception = False )

  # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
  # Post-processing:
  # ----------------

  assert data_config.dataset_dir.exists()


  data_config.is_torchvision_catered_ds = \
    any( [ data_config.dataset in ds[0] for ds in TORCHVISION_CATERED_DATASETS.values() ] )


  # Optional Validation Set Creation based on user input:
  imgs_dir = data_config.dataset_dir
  if not data_config.is_torchvision_catered_ds:
    if ( data_config.dataset_dir/TRAINING_SET_DIRNAME ).exists() and \
       ( data_config.dataset_dir/TRAINING_SET_DIRNAME ).is_dir():
      imgs_dir = data_config.dataset_dir/TRAINING_SET_DIRNAME
    if data_config.include_valid_set:
      imgs_dir = make_valid_set(
                   dataset_dir = data_config.dataset_dir,
                   training_set_dirname = TRAINING_SET_DIRNAME,
                   valid_set_dirname = VALID_SET_DIRNAME,
                   valid_set_pct = data_config.valid_set_pct,
                   valid_set_idxs = data_config.valid_set_idxs
                 )


  # Configure Image Data with user-specified Transforms:
  if USE_PIL_TRANSFORMS:
    if data_config.dataset_downsample_type is 'nearest':
      data_config.dataset_downsample_type = Image.NEAREST
    elif data_config.dataset_downsample_type in ( 'average', 'box', ):
      data_config.dataset_downsample_type = Image.BOX
    elif data_config.dataset_downsample_type is 'bilinear':
      data_config.dataset_downsample_type = Image.BILINEAR

    ds_transforms = [ ]
    if not data_config.is_torchvision_catered_ds:
      ds_ext = get_dataset_img_extension( imgs_dir )
      if ds_ext not in IMG_EXTENSIONS:  # if `True`, then not automatically converted to PIL Image
        ds_transforms.append( transforms.ToPILImage( ) )
      data_config.ds_ext = ds_ext
    if data_config.center_crop_dim is not None:
      ds_transforms.append( transforms.CenterCrop( data_config.center_crop_dim ) )
    if data_config.random_crop_dim is not None:
      ds_transforms.append( transforms.RandomCrop( data_config.random_crop_dim ) )
    if config is not None:
      if config.model == 'ResNet GAN':
        ds_transforms.append( transforms.Resize( size = ( config.res_samples, config.res_samples, ),
                                                 interpolation = data_config.dataset_downsample_type ) )
      elif config.model in ( 'ProGAN', 'StyleGAN', ):
        ds_transforms.append( transforms.Resize( size = ( config.init_res, config.init_res, ),
                                                 interpolation = data_config.dataset_downsample_type ) )
    else:
      ds_transforms.append(
        partial( transforms.Resize, interpolation = data_config.dataset_downsample_type )
      )
    if data_config.enable_mirror_augmentation:  # data is at target resolution after now
      ds_transforms.append( transforms.RandomHorizontalFlip( p = 0.5 ) )
    ds_transforms.append( transforms.ToTensor( ) )  # this normalizes: maps pixel 0 -> 0. and pixel 255 -> 1.
    mean, std = normalize_center_standardize(
                  dataset = data_config.dataset,
                  normalize = False,  # transforms.ToTensor( ) above already takes care of normalization
                  standardize = data_config.standardize
                )
    ds_transforms.append( transforms.Normalize( mean = mean, std = std ) )
    data_config.ds_transforms = transforms.Compose( ds_transforms )
    data_config.ds_mean = mean
    data_config.ds_std = std
  else:
    raise NotImplementedError( 'Transforms on training data that do not use PIL are not yet implemented.' )

  # print( data_config )

  configs_dir = str( os.path.abspath( os.path.dirname( __file__ ) ) )

  with open( str( Path.home()/'.configs_dir.txt' ), 'wb' ) as f:
    f.write( configs_dir.encode() )

  with open( configs_dir + '/.data_config.p', 'wb' ) as f:
    pickle.dump( data_config, f, protocol = 3 )


# -*- coding: UTF-8 -*-

"""Custom layers.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

from .initializer import Initializer

import torch
from torch import nn
import torch.nn.functional as F

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Lambda:
# -------

class Lambda( nn.Module ):
  """Converts any function into a PyTorch Module."""
  def __init__( self, func, **kwargs ):
    super( Lambda, self ).__init__()
    self.func = func
    if kwargs:
      self.kwargs = kwargs
    else:
      self.kwargs = {}

  def forward( self, x ):
    return self.func( x, **self.kwargs )

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Blur (Low-pass Filtering):
# --------------------------

# TODO: Add More (such as Gaussian Blur)
def get_blur_op( blur_type, num_channels ):
  """Options for low-pass filter operations. Only 3x3 kernels supported currently."""
  if blur_type.casefold() == 'box':
    blur_filter = torch.FloatTensor( [ 1./9 ] ).expand( num_channels, 1, 3, 3 )
    stride = 3
  elif blur_type.casefold() == 'binomial':
    blur_filter = torch.FloatTensor( [ [ 1./16, 2./16, 1./16 ],
                                       [ 2./16, 4./16, 2./16 ],
                                       [ 1./16, 2./16, 1./16 ] ] ).expand( num_channels, 1, 3, 3 )
    stride = 1
  elif blur_type.casefold() == 'gaussian':
    raise NotImplementedError( 'Gaussian blur not yet implemented.' )

  # meant to work as a PyTorch Module
  blur_op = Lambda( lambda x: F.conv2d( x, blur_filter.to( x.device ),
                    stride = stride, padding = 1, groups = num_channels ) )

  return blur_op

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Pooling:
# --------

class NearestPool2d( nn.Module ):
  def __init__( self ):
    super( NearestPool2d, self ).__init__()
  def forward( self, x ):
    if x.dim() == 3:
      x = x.view( -1, *x.shape )
    return F.interpolate( x, scale_factor = .5, mode = 'nearest' )

class BilinearPool2d( nn.Module ):
  def __init__( self, align_corners ):
    super( BilinearPool2d, self ).__init__()
    self.align_corners = align_corners
  def forward( self, x ):
    if x.dim() == 3:
      x = x.view( -1, *x.shape )
    return F.interpolate( x, scale_factor = .5, mode = 'bilinear',
                          align_corners = self.align_corners )

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Normalization:
# --------------

class PixelNorm2d( nn.Module ):
  def __init__( self ):
    super( PixelNorm2d, self ).__init__()

  def forward( self, x, eps = 1.e-8 ):
    return x * ( ( ( x**2 ).mean( dim = 1, keepdim = True ) + eps ).rsqrt() )

class NormalizeLayer( nn.Module ):
  """All normalization methods in one place."""
  # TODO: Implement Spectral Normalization
  def __init__( self, norm_type, ni = None, res = None ):
    super( NormalizeLayer, self ).__init__()

    norm_type = norm_type.lower( )

    if norm_type in ( 'pixelnorm', 'pixel norm', ):
      self.norm = PixelNorm2d( )
    elif norm_type in ( 'instancenorm', 'instance norm', ):
      self.norm = nn.InstanceNorm2d( ni, eps = 1.e-8 )
    elif norm_type in ( 'batchnorm', 'batch norm', ):
      assert isinstance( ni, int )
      self.norm = nn.BatchNorm2d( ni )
    elif norm_type in ( 'layernorm', 'layer norm', ):
      # self.norm = F.layer_norm
      assert isinstance( ni, int ); assert isinstance( res, int )
      self.norm = nn.LayerNorm( [ ni, res, res ] )
    else:
      raise Exception( f'`norm_type` == "{norm_type}" not supported.' )

  def forward( self, x ):
      return self.norm( x )

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Minibatch Standard Deviation:
# -----------------------------

def concat_mbstd_layer( x, group_size = 4 ):
  """Minibatch Standard Deviation layer."""
  _sz = x.size()

  # `group_size` must be less than or equal to and divisible into minibatch size;
  # otherwise use the entire minibatch
  group_size = min( _sz[0], group_size )
  if _sz[0] % group_size != 0:
    group_size = _sz[0]
  G = _sz[0] // group_size

  if group_size > 1:
    mbstd_map = x.view( G, group_size, _sz[1], _sz[2], _sz[3] )
    mbstd_map = torch.var( mbstd_map, dim = 1 )
    mbstd_map = torch.sqrt( mbstd_map + 1.e-8 )
    mbstd_map = mbstd_map.view( G, -1 )
    mbstd_map = torch.mean( mbstd_map, dim = 1 ).view( G, 1 )
    mbstd_map = mbstd_map.expand( G, _sz[2]*_sz[3] ).view( ( G, 1, 1, _sz[2], _sz[3] ) )
    mbstd_map = mbstd_map.expand( G, group_size, -1, -1, -1 )
    mbstd_map = mbstd_map.contiguous().view( ( -1, 1, _sz[2], _sz[3] ) )
  else:
    mbstd_map = torch.zeros( _sz[0], 1, _sz[2], _sz[3], device = x.device )

  return torch.cat( ( x, mbstd_map, ), dim = 1 )

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Bread-and-butter Linear Operations, but with features such as
# equalized LR, custom initialization, etc:
# ------------------------------------------------------------

class Conv2dEx( nn.Module ):
  def __init__( self, ni, nf, ks, stride = 1, padding = 0, groups = 1, init = 'he', init_type = 'default',
                gain_sq_base = 2., equalized_lr = False, lrmul = 1., include_bias = True ):
    super( Conv2dEx, self ).__init__()

    self.ni = ni; self.nf = nf

    init = init.casefold()
    init_type = init_type.casefold()
    self.equalized_lr = equalized_lr
    self.use_lrmul = True if lrmul != 1. else False
    self.lrmul = lrmul

    self.initializer = None
    if init_type != 'standard normal':
      self.initializer = Initializer( init = init, init_type = init_type,
                                      gain_sq_base = gain_sq_base,
                                      equalized_lr = equalized_lr )

    self.conv2d = nn.Conv2d(
      ni, nf, kernel_size = ks, stride = stride,
      padding = padding, groups = groups, bias = include_bias
    )

    # Weights:
    self.wscale = None
    if init_type == ( 'default', 'resnet', ) and init is not None:
      bound = self.initializer.get_init_bound_layer(
        tensor = self.conv2d.weight,
        distribution_type = 'Uniform', stride = stride
      )
      if equalized_lr:
        self.wscale = bound
        self.conv2d.weight.data.uniform_( -1. / lrmul, 1. / lrmul )
      else:
        self.conv2d.weight.data.uniform_( -bound / lrmul, bound / lrmul )
    elif init_type in ( 'progan', 'stylegan', ) and init is not None:
      bound = self.initializer.get_init_bound_layer(
        tensor = self.conv2d.weight,
        distribution_type = 'Normal', stride = stride
      )
      if equalized_lr:
        self.wscale = bound
        self.conv2d.weight.data.normal_( 0., 1. / lrmul )
      else:
        self.conv2d.weight.data.normal_( 0., bound / lrmul )
    elif init_type == 'standard normal' and init is None and not equalized_lr and not self.use_lrmul:
      self.conv2d.weight.data.normal_( 0., 1. )

    # Biases:
    self.bias = None
    if include_bias:
      # init biases as 0, according to official implementations.
      self.conv2d.bias.data.fill_( 0 )

  def forward( self, x ):
    if self.equalized_lr:
      x = self.conv2d( x.mul( self.wscale ) )
    else:
      x = self.conv2d( x )

    if self.use_lrmul:
      x *= self.lrmul

    return x

class Conv2dBias( nn.Module ):
  def __init__( self, nf, lrmul = 1., device = 'cuda' if torch.cuda.is_available() else 'cpu' ):
    super( Conv2dBias, self ).__init__()

    self.use_lrmul = True if lrmul != 1. else False
    self.lrmul = lrmul

    self.bias = nn.Parameter( torch.FloatTensor( 1, nf, 1, 1 ).to( device ).fill_( 0 ) )

  def forward( self, x ):
    if self.use_lrmul:
      return x + self.bias.mul( self.lrmul )
    else:
      return x + self.bias

# ............................................................................ #

class LinearEx( nn.Module ):
  def __init__( self, nin_feat, nout_feat, init = 'xavier', init_type = 'default',
                gain_sq_base = 2., equalized_lr = False, lrmul = 1., include_bias = True ):
    super( LinearEx, self ).__init__()

    self.nin_feat = nin_feat; self.nout_feat = nout_feat

    init = init.casefold()
    init_type = init_type.casefold()
    self.equalized_lr = equalized_lr
    self.use_lrmul = True if lrmul != 1. else False
    self.lrmul = lrmul

    self.initializer = None
    if init_type != 'standard normal':
      self.initializer = Initializer( init = init, init_type = init_type,
                                      gain_sq_base = gain_sq_base,
                                      equalized_lr = equalized_lr )

    self.linear = nn.Linear( nin_feat, nout_feat, bias = include_bias )

    # Weights:
    self.wscale = None
    if init_type in ( 'default', 'resnet', ) and init is not None:
      bound = self.initializer.get_init_bound_layer(
        tensor = self.linear.weight,
        distribution_type = 'Uniform'
      )
      if equalized_lr:
        self.wscale = bound
        self.linear.weight.data.uniform_( -1. / lrmul, 1. / lrmul )
      else:
        self.linear.weight.data.uniform_( -bound / lrmul, bound / lrmul )
    elif init_type in ( 'progan', 'stylegan', ) and init is not None:
      bound = self.initializer.get_init_bound_layer(
        tensor = self.linear.weight,
        distribution_type = 'Normal'
      )
      if equalized_lr:
        self.wscale = bound
        self.linear.weight.data.normal_( 0., 1. / lrmul )
      else:
        self.linear.weight.data.normal_( 0., bound / lrmul )
    elif init_type == 'standard normal' and init is None and not equalized_lr and not self.use_lrmul:
      self.linear.weight.data.normal_( 0., 1. )

    # Biases:
    self.bias = None
    if include_bias:
      # init biases as 0, according to official implementations.
      self.linear.bias.data.fill_( 0 )

  def forward( self, x ):
    if self.equalized_lr:
      x = self.linear( x.mul( self.wscale ) )
    else:
      x = self.linear( x )

    if self.use_lrmul:
      x *= self.lrmul

    return x

class LinearBias( nn.Module ):
  def __init__( self, nout_feat, lrmul = 1., device = 'cuda' if torch.cuda.is_available() else 'cpu' ):
    super( LinearBias, self ).__init__()

    self.use_lrmul = True if lrmul != 1. else False
    self.lrmul = lrmul

    self.bias = nn.Parameter( torch.FloatTensor( 1, nout_feat ).to( device ).fill_( 0 ) )

  def forward( self, x ):
    if self.use_lrmul:
      return x + self.bias.mul( self.lrmul )
    else:
      return x + self.bias


# -*- coding: UTF-8 -*-

"""Backpropagation utility functions.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

from _int import FMAP_SAMPLES, SUPPORTED_ARCHS

from functools import partial

import torch
import torch.nn.functional as F

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Loss Functions:
# ---------------

def wasserstein_distance_gen( outb ):
  # return outb.mean().abs()
  return -( outb.mean() )

def nonsaturating_loss_gen( outb ):
  return F.binary_cross_entropy_with_logits( input = outb,
                                             target = torch.ones( *( outb.shape ) ).to( outb.device ),
                                             reduction = 'mean' )

def minimax_loss_gen( outb ):
  return -F.binary_cross_entropy_with_logits( input = outb,
                                              target = torch.zeros( *( outb.shape ) ).to( outb.device ),
                                              reduction = 'mean' )

def wasserstein_distance_disc( outb, yb ):
  """Assumes you sample a pair of real & fake each time."""
  # return -( ( outb - yb ).mean().abs() )
  return ( outb - yb ).mean()

def minimax_loss_disc( outb, yb ):
  return F.binary_cross_entropy_with_logits( input = outb,
                                             target = torch.zeros( *( outb.shape ) ).to( outb.device ),
                                             reduction = 'mean' ) + \
         F.binary_cross_entropy_with_logits( input = yb,
                                             target = torch.ones( *( outb.shape ) ).to( outb.device ),
                                             reduction = 'mean' )

def wasserstein_distance_with_gp_disc( outb, yb, nn_disc, gp_type = 'wgan-gp', lda = 10., gamma = 1. ):
  """Assumes you sample a pair of real & fake each time."""
  # return -( ( outb - yb ).mean().abs() ) + calc_gp( gp_type, nn_disc, outb, yb, lda, gamma )
  return ( outb - yb ).mean() + calc_gp( gp_type, nn_disc, outb, yb, lda, gamma )

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Gradient Regularizers:
# ----------------------

def calc_gp( gp_type, nn_disc, gen_data, real_data, lda = 10., gamma = 1. ):
  """Gradient penalty with custom lambda and gamma values."""
  gp_type = gp_type.casefold()

  if gp_type in ( 'wgan-gp', 'r1', ):
    real_data = real_data.view(
      -1, FMAP_SAMPLES, real_data.shape[2], real_data.shape[3]
    )
  if gp_type in ( 'wgan-gp', 'r2', ):
    gen_data = gen_data.view(
      -1, FMAP_SAMPLES, gen_data.shape[2], gen_data.shape[3]
    )

  # whether to penalize the discriminator gradients on the real distribution, fake distribution, or an interpolation of both
  # TODO: Implement class-conditioning (in the discriminator) option (Mirza & Osindero, 2014).
  if gp_type == 'wgan-gp':
    eps = torch.rand( gen_data.shape[0], 1, 1, 1, device = gen_data.device )
    xb = ( eps * gen_data.detach() + \
         ( 1 - eps ) * real_data.detach() ).to( gen_data.device )
  elif gp_type == 'r1':
    xb = real_data.detach().to( real_data.device )
  elif gp_type == 'r2':
    xb = gen_data.detach().to( gen_data.device )

  # now start tracking in the computation graph again
  xb.requires_grad_( True )

  if nn_disc.__class__.__name__ in SUPPORTED_ARCHS[ 'AC Discriminators' ]:
    outb, _ = nn_disc( xb )
  elif nn_disc.__class__.__name__ in SUPPORTED_ARCHS[ 'Discriminators' ]:
    outb = nn_disc( xb )
  # print( xb.requires_grad, outb.requires_grad )
  outb_grads = torch.autograd.grad(
    outb,
    xb,
    grad_outputs = torch.ones( gen_data.shape[0] ).to( gen_data.device ),
    create_graph = True, retain_graph = True, only_inputs = True )[0]

  if gp_type == 'wgan-gp':
    if gamma != 1.:
      outb_gp = \
        ( ( outb_grads.norm( 2, dim = 1 ) - gamma )**2 / gamma**2 ).mean() * lda
    else:
      outb_gp = \
        ( ( outb_grads.norm( 2, dim = 1 ) - 1. )**2 ).mean() * lda
  elif gp_type in ( 'r1', 'r2', ):
    outb_gp = ( outb_grads.norm( 2, dim = 1 )**2 ).mean() * lda / 2.

  return outb_gp

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
# Optimizer:
# ----------

def configure_adam_for_gan( lr_base, betas:tuple, eps = 1.e-8, wd = 0 ):
  assert isinstance( betas, tuple )

  adam_gan = partial(
    torch.optim.Adam,
    lr = lr_base,
    betas = betas,
    eps = eps,
    weight_decay = wd
  )

  return adam_gan


# -*- coding: UTF-8 -*-

"""Learner for StyleGANs.

  Typical usage example:

  First configure your desired GAN on the command-line:
    go to root directory...
    $ python config.py stylegan
    $ python data_config.py FFHQ path/to/datasets/ffhq

  Then write a custom script (or use train.py):
    from gan_lab import get_current_configuration
    from gan_lab.utils.data_utils import prepare_dataset, prepare_dataloader
    from gan_lab.stylegan.learner import StyleGANLearner

    # get most recent configurations:
    config = get_current_configuration( 'config' )
    data_config = get_current_configuration( 'data_config' )

    # get DataLoader(s)
    train_ds, valid_ds = prepare_dataset( data_config )
    train_dl, valid_dl, z_valid_dl = prepare_dataloader( config, data_config, train_ds, valid_ds )

    # instantiate StyleGANLearner and train:
    learner = StyleGANLearner( config )
    learner.train( train_dl, valid_dl, z_valid_dl )   # train for config.num_main_iters iterations
    learner.config.num_main_iters = 300000            # this is one example of changing your instantiated learner's configurations
    learner.train( train_dl, valid_dl, z_valid_dl )   # train for another 300000 iterations

Note that the above custom script is just a more flexible alternative to running
train.py (you can, for example, run the above on a Jupyter Notebook). You can
always just run train.py.
"""

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

from .base import StyleGAN
from .architectures import StyleAddNoise, StyleGenerator
from _int import get_current_configuration, LearnerConfigCopy
from progan.architectures import ProDiscriminator
from progan.learner import ProGANLearner
from utils.latent_utils import gen_rand_latent_vars
from utils.backprop_utils import calc_gp, configure_adam_for_gan
from utils.custom_layers import Conv2dBias, LinearBias

import os
import sys
from abc import ABC
import copy
import logging
import warnings
from pathlib import Path
from functools import partial
from itertools import accumulate
from timeit import default_timer as timer

import numpy as np
from PIL import Image
from indexed import IndexedOrderedDict
import matplotlib.pyplot as plt
plt.rcParams.update( { 'figure.max_open_warning': 0 } )
import torch
from torch import nn
import torch.nn.functional as F
from torchvision import transforms

# from tqdm import tqdm

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

NONREDEFINABLE_ATTRS = ( 'model', 'init_res', 'res_samples', 'res_dataset', 'len_latent',
                         'num_classes', 'class_condition', 'use_auxiliary_classifier',
                         'model_upsample_type', 'model_downsample_type', 'align_corners',
                         'blur_type', 'nonlinearity', 'use_equalized_lr', 'normalize_z',
                         'use_pixelnorm', 'mbstd_group_size', 'use_ewma_gen',
                         'use_instancenorm', 'use_noise', 'pct_mixing_reg',
                         'beta_trunc_trick', 'psi_trunc_trick', 'cutoff_trunc_trick',
                         'len_dlatent', 'mapping_num_fcs', 'mapping_lrmul', )

REDEFINABLE_FROM_LEARNER_ATTRS = ( 'batch_size', 'loss', 'gradient_penalty',
                                   'optimizer', 'lr_sched', 'latent_distribution', )

COMPUTE_EWMA_VIA_HALFLIFE = True
EWMA_SMOOTHING_HALFLIFE = 10.
EWMA_SMOOTHING_BETA = .999

# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

class StyleGANLearner( ProGANLearner ):
  """GAN Learner specifically designed for StyleGAN architectures.

  Once instantiated, the StyleGANLearner object's configuration can be changed, but only via
  its self.config attribute (i.e. running 'python config.py [model]' post-instantiation
  will not affect this learner's configuration).
  """
  def __init__( self, config ):
    super( StyleGANLearner, self ).__init__( config )

    if self.model == 'StyleGAN':

      # If you want to change an attribute in an already-instantiated StyleGANLearner's config or data_config,
      # change self.config (below) instead of config and self.data_config (also below) instead of data_config:
      self.config = LearnerConfigCopy( config,
                                       self.__class__.__name__,
                                       NONREDEFINABLE_ATTRS,
                                       REDEFINABLE_FROM_LEARNER_ATTRS )
      # pretrained models loaded later on for evaluation should not require data_config.py to have been run:
      self._is_data_configed = False; self._stats_set = False
      self._update_data_config( raise_exception = False )

      self._latent_distribution = self.config.latent_distribution

      global StyleGAN, StyleGenerator, ProDiscriminator, StyleDiscriminator
      StyleGAN = type( 'StyleGAN', ( nn.Module, ABC, ), dict( StyleGAN.__dict__ ) )
      StyleGAN.reset_state( )

      StyleGenerator = type( 'StyleGenerator', ( StyleGAN, ), dict( StyleGenerator.__dict__ ) )
      self.gen_model = StyleGenerator(
        final_res = self.config.res_samples,
        latent_distribution = self._latent_distribution,
        len_latent = self.config.len_latent,
        len_dlatent = self.config.len_dlatent,
        mapping_num_fcs = self.config.mapping_num_fcs,
        mapping_lrmul = self.config.mapping_lrmul,
        use_instancenorm = self.config.use_instancenorm,
        use_noise = self.config.use_noise,
        upsampler = self.gen_model_upsampler,
        blur_type = self.config.blur_type,
        nl = self.nl,
        num_classes = self.num_classes_gen,
        equalized_lr = self.config.use_equalized_lr,
        normalize_z = self.config.normalize_z,
        use_pixelnorm = self.config.use_pixelnorm,
        pct_mixing_reg = self.config.pct_mixing_reg,
        truncation_trick_params = { 'beta': self.config.beta_trunc_trick,
                                    'psi': self.config.psi_trunc_trick,
                                    'cutoff_stage': self.config.cutoff_trunc_trick }
      )

      # Create `StyleDiscriminator` type object by converting `ProDiscriminator` class into `StyleDiscriminator` class:
      StyleDiscriminator = type( 'StyleDiscriminator', ( StyleGAN, ), dict( ProDiscriminator.__dict__ ) )
      self.disc_model = StyleDiscriminator(
        final_res = self.config.res_samples,
        pooler = self.disc_model_downsampler,
        blur_type = self.config.blur_type,
        nl = self.nl,
        num_classes = self.num_classes_disc,
        equalized_lr = self.config.use_equalized_lr,
        mbstd_group_size = self.config.mbstd_group_size
      )

      # If one wants to start at a higher resolution than 4:
      assert self.config.init_res <= self.config.res_samples
      if self.config.init_res > 4:
        _init_res_log2 = int( np.log2( self.config.init_res ) )
        if float( self.config.init_res ) != 2**_init_res_log2:
          raise ValueError( 'Only resolutions that are powers of 2 are supported.' )
        num_scale_inc = _init_res_log2 - 2
        for _ in range( num_scale_inc ):
          self.gen_model.increase_scale()
          self.disc_model.increase_scale()
        self.gen_model.fade_in_phase = False  # this applies it to both networks simultaneously

      # Generator and Discriminator state data must match:
      assert self.gen_model.cls_base.__dict__ == \
             self.disc_model.cls_base.__dict__

      # Initialize EWMA Generator Model:
      self.gen_model_lagged = None
      if self.config.use_ewma_gen:
        _orig_mode = self.gen_model.training
        self.gen_model.train()
        self.gen_model.to( 'cpu' )
        with torch.no_grad():
          self.gen_model_lagged = copy.deepcopy( self.gen_model )  # for memory efficiency in GPU
          self.gen_model_lagged.to( self.config.metrics_dev )
          self.gen_model_lagged.train( mode = _orig_mode )
        self.gen_model.train( mode = _orig_mode )

      self.gen_model.to( self.config.dev )
      self.disc_model.to( self.config.dev )

      self.batch_size = self.config.bs_dict[ self.gen_model.curr_res ]

      if self.cond_gen:
        self.labels_one_hot_disc = self._tensor( self.batch_size, self.num_classes )
        self.labels_one_hot_gen = self._tensor( self.batch_size * self.config.gen_bs_mult, self.num_classes )

      # Loss Function:
      self._loss = config.loss.casefold()
      self._set_loss( )

      # Gradient Regularizer:
      self.gp_func = partial(
        calc_gp,
        gp_type = self._gradient_penalty,
        nn_disc = self.disc_model,
        lda = self.config.lda,
        gamma = self.config.gamma
      )

      # Optimizer:
      self._set_optimizer( )

      # Epsilon Loss to punish possible outliers from training distribution:
      self.eps = False
      if self.config.eps_drift > 0:
        self.eps = True

      # Print configuration:
      print( '-------- Initialized Model Configuration --------' )
      print( self.config )
      print( '-------------------------------------------------' )
      # print( "  If you would like to alter any of the above configurations,\n" + \
      #        "  please do so via altering your instantiated StyleGANLearner().config's attributes." )
      print( '\n    Ready to train!\n' )

  def _apply_lagged_weights( self, m ):
    # TODO: Include support for other learnable layers such as BatchNorm
    _keys = m.state_dict().keys()
    if isinstance( m, ( nn.Linear, nn.Conv2d, LinearBias, Conv2dBias, ) ):
      if 'weight' in _keys:
        m.weight = nn.Parameter( self.lagged_params.values()[ self._param_tensor_num + 1 ] )
        self._param_tensor_num += 1
      if 'bias' in _keys:
        m.bias = nn.Parameter( self.lagged_params.values()[ self._param_tensor_num + 1 ] )
        self._param_tensor_num += 1
    elif isinstance( m, StyleAddNoise ):
      if 'noise_weight' in _keys:
        m.noise_weight = nn.Parameter( self.lagged_params.values()[ self._param_tensor_num + 1 ] )
        self._param_tensor_num += 1
    elif isinstance( m, StyleGenerator ):
      if 'const_input' in _keys:
        m.const_input = nn.Parameter( self.lagged_params.values()[ 0 ] )
        self._param_tensor_num += 1

  @property
  def latent_distribution( self ):
    return self._latent_distribution

  @latent_distribution.setter
  def latent_distribution( self, new_latent_distribution ):
    new_latent_distribution = new_latent_distribution.casefold()
    self._latent_distribution = new_latent_distribution
    self.gen_model.latent_distribution = new_latent_distribution
    if self.config.use_ewma_gen:
      self.gen_model_lagged.latent_distribution = new_latent_distribution

  # def reset_stylegan_state( self ):
  #   self.gen_model.cls_base.reset_state( )  # this applies to both networks simultaneously

  # .......................................................................... #

  @torch.no_grad()
  def plot_sample( self, z_test, z_mixing = None, style_mixing_stage = None, noise = None, label = None, time_average = True ):
    """Plots and shows 1 sample from input latent code; offers stylemixing and noise input capabilities."""
    if self.ds_mean is None or self.ds_std is None:
      raise ValueError( "This model does not hold any information about your dataset's mean and/or std.\n" + \
                        "Please provide these (either from your current data configuration or from your pretrained model)." )

    for z in ( z_test, z_mixing, ):
      if z is not None:
        if z.dim() == 2:
          if z.shape[0] != 1:
            raise IndexError( 'This method only permits plotting 1 generated sample at a time.' )
        elif z.dim() != 1:
          raise IndexError( 'Incorrect dimensions of input latent vector. Must be either `dim == 1` or `dim == 2`.' )
        if not self.cond_gen:
          if z.shape[-1] != self.config.len_latent:
            message = f"Input latent vector must be of size {self.config.len_latent}."
            raise IndexError( message )
        else:
          if z.shape[-1] != self.config.len_latent + self.num_classes_gen:
            message = f"This is a generator class-conditioned model. So please make sure to append a one-hot encoded vector\n" + \
                      f"of size {self.num_classes_gen} that indicates the to-be generated sample's class to a latent vector of\n" + \
                      f"size {self.config.len_latent}. Total input size must therefore be {self.config.len_latent + self.num_classes_gen}."
            raise IndexError( message )

    # z_test = z_test.to( self.config.dev )
    x_test = \
      self.gen_model_lagged(
        z_test,
        x_mixing = z_mixing,
        style_mixing_stage = style_mixing_stage,
        noise = noise ).squeeze() if time_average else \
      self.gen_model(
        z_test,
        x_mixing = z_mixing,
        style_mixing_stage = style_mixing_stage,
        noise = noise ).squeeze()

    if label is not None:
      print( f'Label Index for Generated Image: {label}' )

    logger = logging.getLogger()
    _old_level = logger.level
    logger.setLevel( 100 )  # ignores potential "clipping input data" warning
    plt.imshow( ( ( ( x_test ) \
                          .cpu().detach() * self.ds_std ) + self.ds_mean ) \
                          .numpy().transpose( 1, 2, 0 ), interpolation = 'none'
    )
    logger.setLevel( _old_level )
    plt.show()

  @torch.no_grad()
  def make_stylemixing_grid( self, zs_sourceb, zs_coarse = [], zs_middle = [], zs_fine = [], labels = None, time_average = True, save_path = None ):
    """Generates style-mixed grid of images, emulating Figure 3 in Karras et al. 2019."""
    assert any( len( zs ) for zs in ( zs_coarse, zs_middle, zs_fine, ) )
    if self.ds_mean is None or self.ds_std is None:
      raise ValueError( "This model does not hold any information about your dataset's mean and/or std.\n" + \
                        "Please provide these (either from your current data configuration or from your pretrained model)." )

    szs = []
    stages = [ 0 ]
    for m, zs in enumerate( ( zs_sourceb, zs_coarse, zs_middle, zs_fine, ) ):
      szs.append( 1 if isinstance( zs, torch.Tensor ) and zs.dim() == 1 else len( zs ) )
      if szs[m]:
        if zs.dim() > 2:
          raise IndexError( 'Incorrect dimensions of input latent vector. Must be either `dim == 1` or `dim == 2`.' )
        else:
          if m == 1:
            stage = 1  # stage for style-mixing of coarse features from source b
            if zs.dim() == 1: zs_coarse.unsqueeze_( dim = 0 )
          elif m == 2:
            stage = 4  # stage for style-mixing of mid-level features from source b
            if zs.dim() == 1: zs_middle.unsqueeze_( dim = 0 )
          elif m == 3:
            stage = 8  # stage for style-mixing of fine-grained features from source b
            if zs.dim() == 1: zs_fine.unsqueeze_( dim = 0 )
          if m: stages.append( stage )
        if not self.cond_gen:
          if zs.shape[1] != self.config.len_latent:
            message = f"Input latent vector must be of size {self.config.len_latent}."
            raise IndexError( message )
        else:
          if zs.shape[1] != self.config.len_latent + self.num_classes_gen:
            message = f"This is a generator class-conditioned model. So please make sure to append a one-hot encoded vector\n" + \
                      f"of size {self.num_classes_gen} that indicates the to-be generated sample's class to a latent vector of\n" + \
                      f"size {self.config.len_latent}. Total input size must therefore be {self.config.len_latent + self.num_classes_gen}."
            raise IndexError( message )
      else:
          stages.append( 0 )

    ncols_tot = 1 + szs[0]; nrows_tot = 1 + sum( szs[1:] )
    cum_szs = [0] + list( accumulate( szs[1:] ) )
    fig = plt.figure( figsize = ( 8. * ( ncols_tot / nrows_tot ), 8. if labels is None else 9., ), facecolor = 'white' )
    axs = fig.subplots( ncols = ncols_tot, nrows = nrows_tot )
    fig.tight_layout( pad = 0 )
    _fctrs = ( ( fig.subplotpars.wspace / 0.47267497603068315 ), ( fig.subplotpars.hspace / 0.28000000000000086 ), )
    gaps = ( ( .002 * _fctrs[0] * ( fig.dpi / fig.get_size_inches().mean() ) ), \
              ( .002 * _fctrs[1] * ( fig.dpi / fig.get_size_inches()[1] ) ), )
    # wspace = fig.subplotpars.wspace
    # hspace = fig.subplotpars.hspace
    wspace = plt.rcParams[ 'figure.subplot.wspace' ] * _fctrs[0]
    hspace = plt.rcParams[ 'figure.subplot.hspace' ] * _fctrs[1]
    start = 0
    logger = logging.getLogger(); _old_level = logger.level; logger.setLevel( 100 )  # ignores potential "clipping input data" warning
    for m, zs in enumerate( ( zs_sourceb, zs_coarse, zs_middle, zs_fine, ) ):
      if szs[m]:
        nrows = 1 if not m else szs[ m ]
        if m: start = row + 1
        for row in range( start, start + nrows ):
          for col in range( ncols_tot ):
            axs[row][col].axis( 'off' )
            axs[row][col].set_aspect( 'equal' )
            l, b, w, h = axs[row][col].get_position().bounds
            l *= ( 1. - wspace ); b *= ( 1. - hspace ); w *= ( 1. + wspace ); h *= ( 1. + hspace )
            if m > 1: b -= ( m - 1 ) *gaps[1]
            if not row and not col:
              axs[row][col].set_position( [ l - gaps[0], b + gaps[1], w, h ] )
            else:
              if ( not row ) != ( not col ):
                if row:
                  axs[row][col].set_position( [ l - gaps[0], b, w, h ] )
                  if row == 1:
                    axs[row][col].set_title( 'Source A', fontweight = 'bold' )
                elif col:
                  axs[row][col].set_position( [ l, b + gaps[1], w, h ] )
                  if col == 1:
                    axs[row][col].axis( 'on' )
                    axs[row][col].spines['top'].set_visible(False)
                    axs[row][col].spines['right'].set_visible(False)
                    axs[row][col].spines['bottom'].set_visible(False)
                    axs[row][col].spines['left'].set_visible(False)
                    axs[row][col].set_xticks( [] )
                    axs[row][col].set_yticks( [] )
                    axs[row][col].set_ylabel( 'Source B', fontweight = 'bold', fontsize = plt.rcParams[ 'axes.titlesize' ] )
                x = self.gen_model_lagged( zs_sourceb[ col - 1 ] if not m else zs[ row - cum_szs[m-1] - 1 ] ).squeeze() if time_average else \
                    self.gen_model( zs_sourceb[ col - 1 ] if not m else zs[ row - cum_szs[m-1] - 1 ] ).squeeze()
              elif row and col:
                axs[row][col].set_position( [ l, b, w, h ] )
                x = self.gen_model_lagged( zs[ row - cum_szs[m-1] - 1 ], x_mixing = zs_sourceb[ col - 1 ], style_mixing_stage = stages[ m ] ).squeeze() if time_average else \
                    self.gen_model( zs[ row - cum_szs[m-1] - 1 ], x_mixing = zs_sourceb[ col - 1 ], style_mixing_stage = stages[ m ] ).squeeze()
              axs[row][col].imshow(
                ( ( x.cpu().detach() * self.ds_std ) + self.ds_mean ).numpy().transpose( 1, 2, 0 ), interpolation = 'none'
              )
              if labels is not None:
                axs[row][col].set_title( str( labels[ row*ncols_tot + col ].item() ) )

    # center the axes of each subplot on the figure after moving everything around
    fig_top_right = fig.get_window_extent().corners()[3]
    l,b,w,h = axs[0,-1].get_position().bounds; top_right = ( l + w, b + h, )
    l,b,w,h = axs[-1,0].get_position().bounds; bot_left = ( l, b, )
    shift = tuple( -bot_left[i] + ( 1. - top_right[i] ) / 2. for i in range( 2 ) )
    for row in range( nrows_tot ):
      for col in range( ncols_tot ):
        l, b, w, h = axs[row][col].get_position().bounds
        axs[row][col].set_position( [ l + shift[0], b + shift[1], w, h ] )

    # ylabel for Coarse, Middle, and Fine styles from Source B
    for k in range( 1, m + 1 ):
      if szs[k]:
        l, b, w, h = axs[cum_szs[k]][0].get_position().bounds
        center_y = ( szs[k] * h ) / 2.
        if k == 1: ylabel = 'Coarse'
        elif k == 2: ylabel = 'Middle'
        elif k == 3: ylabel = 'Fine'
        ylabel += ' from B' if szs[k] < 2 else ' styles from Source B'
        fig.text( l - gaps[0], b + center_y, ylabel, va = 'center', rotation = 'vertical' )
    logger.setLevel( _old_level )

    # maintain resolution of images and save
    if save_path is not None:
      bbox = axs[-1][-1].get_window_extent().transformed( fig.dpi_scale_trans.inverted() )
      dpi = ( self.gen_model_lagged.curr_res if time_average else self.gen_model.curr_res ) / bbox.height
      fig.savefig( save_path, dpi = dpi, bbox_inches = 'tight', pad_inches = 0, facecolor = fig.get_facecolor() )
      
    return ( fig, axs, )

  # .......................................................................... #

  def save_model( self, save_path:Path ):
    if self.not_trained_yet:
      raise Exception( 'Please train your model for atleast 1 iteration before saving.' )

    warnings.filterwarnings( 'ignore', category = UserWarning )

    self.gen_model_metadata = { 'gen_model_upsampler': self.gen_model_upsampler,
                                'num_classes_gen': self.num_classes_gen }
    self.disc_model_metadata = { 'disc_model_downsampler': self.disc_model_downsampler,
                                 'num_classes_disc': self.num_classes_disc }

    scheduler_state_dict_save_args = {
      'scheduler_gen_state_dict': self.scheduler_gen.state_dict() if self.sched_bool else None,
      'scheduler_disc_state_dict': self.scheduler_disc.state_dict() if self.sched_bool else None
    }

    save_path = Path( save_path )
    save_path.parents[0].mkdir( parents = True, exist_ok = True )

    torch.save( {
                'config': self.config,
                'curr_res': self.gen_model.curr_res,
                'alpha': self.gen_model.alpha,
                'use_truncation_trick': self.gen_model.use_truncation_trick,
                'trunc_cutoff_stage': self.gen_model.trunc_cutoff_stage,
                'w_eval_psi': self.gen_model.w_eval_psi,
                'w_ewma_beta': self.gen_model.w_ewma_beta,
                'w_ewma': self.gen_model.w_ewma.to( 'cpu' ),
                'w_ewma_lagged': self.gen_model_lagged.w_ewma.to( 'cpu' ) if self.config.use_ewma_gen else None,
                'trained_with_noise': self.gen_model._trained_with_noise,
                'pct_mixing_reg': self.gen_model.pct_mixing_reg,
                'gen_model_metadata': self.gen_model_metadata,
                'gen_model_state_dict': self.gen_model.state_dict(),
                'gen_model_lagged_state_dict': self.gen_model_lagged.state_dict() if self.config.use_ewma_gen else None,
                'disc_model_metadata': self.disc_model_metadata,
                'disc_model_state_dict': self.disc_model.state_dict(),
                'nl': self.nl,
                'sched_stop_step': self.sched_stop_step,
                'lr_sched': self.lr_sched,
                **scheduler_state_dict_save_args,
                'optimizer': self.optimizer,
                'opt_gen_state_dict': self.opt_gen.state_dict(),
                'opt_disc_state_dict': self.opt_disc.state_dict(),
                'loss': self.loss,
                'gradient_penalty': self.gradient_penalty,
                'batch_size': self.batch_size,
                'curr_dataset_batch_num': self.curr_dataset_batch_num,
                'curr_epoch_num': self.curr_epoch_num,
                'tot_num_epochs': self.tot_num_epochs,
                'dataset_sz': self.dataset_sz,
                'ac': self.ac,
                'cond_gen': self.cond_gen,
                'cond_disc': self.cond_disc,
                'valid_z': self.valid_z.to( 'cpu' ),
                'valid_label': self.valid_label,
                'grid_inputs_constructed': self.grid_inputs_constructed,
                'rand_idxs': self.rand_idxs,
                'gen_metrics_num': self.gen_metrics_num,
                'disc_metrics_num': self.disc_metrics_num,
                'curr_img_num': self.curr_img_num,
                'nimg_transition_lst': self.nimg_transition_lst,
                'not_trained_yet': self.not_trained_yet,
                'ds_mean': self.ds_mean,
                'ds_std': self.ds_std,
                'latent_distribution': self.latent_distribution,
                'curr_phase_num': self.curr_phase_num,
                'lagged_params': self.lagged_params,
                'progressively_grow': self.progressively_grow }, save_path
    )

  def load_model( self, load_path, dev_of_saved_model = 'cpu' ):
    dev_of_saved_model = dev_of_saved_model.casefold()
    assert ( dev_of_saved_model == 'cpu' or dev_of_saved_model == 'cuda' )
    _map_location = lambda storage,loc: storage if dev_of_saved_model == 'cpu' else None

    checkpoint = torch.load( load_path, map_location = _map_location )

    self.config = checkpoint[ 'config' ]

    self.nl = checkpoint[ 'nl' ]
    self._latent_distribution = checkpoint[ 'latent_distribution' ]

    # Load pretrained neural networks:
    global StyleGAN, StyleGenerator, ProDiscriminator, StyleDiscriminator
    StyleGAN = type( 'StyleGAN', ( nn.Module, ABC, ), dict( StyleGAN.__dict__ ) )
    StyleGAN.reset_state( )

    StyleGenerator = type( 'StyleGenerator', ( StyleGAN, ), dict( StyleGenerator.__dict__ ) )
    self.gen_model_metadata = checkpoint[ 'gen_model_metadata' ]
    self.gen_model_upsampler = self.gen_model_metadata[ 'gen_model_upsampler' ]
    self.num_classes_gen = self.gen_model_metadata[ 'num_classes_gen' ]
    self.gen_model = StyleGenerator(
      final_res = self.config.res_samples,
      latent_distribution = self._latent_distribution,
      len_latent = self.config.len_latent,
      len_dlatent = self.config.len_dlatent,
      mapping_num_fcs = self.config.mapping_num_fcs,
      mapping_lrmul = self.config.mapping_lrmul,
      use_instancenorm = self.config.use_instancenorm,
      use_noise = self.config.use_noise,
      upsampler = self.gen_model_upsampler,
      blur_type = self.config.blur_type,
      nl = self.nl,
      num_classes = self.num_classes_gen,
      equalized_lr = self.config.use_equalized_lr,
      normalize_z = self.config.normalize_z,
      use_pixelnorm = self.config.use_pixelnorm,
      pct_mixing_reg = self.config.pct_mixing_reg,
      truncation_trick_params = { 'beta': self.config.beta_trunc_trick,
                                  'psi': self.config.psi_trunc_trick,
                                  'cutoff_stage': self.config.cutoff_trunc_trick }
    )
    self.gen_model.w_ewma_beta = checkpoint[ 'w_ewma_beta' ]
    self.gen_model._w_eval_psi = checkpoint[ 'w_eval_psi' ]
    self.gen_model._trunc_cutoff_stage = checkpoint[ 'trunc_cutoff_stage' ]
    self.gen_model.use_truncation_trick = checkpoint[ 'use_truncation_trick' ]
    self.gen_model.w_ewma = checkpoint[ 'w_ewma' ]
    self.gen_model._trained_with_noise = checkpoint[ 'trained_with_noise' ]
    self.gen_model._use_noise = checkpoint[ 'trained_with_noise' ]
    self.gen_model.pct_mixing_reg = checkpoint[ 'pct_mixing_reg' ]
    self.gen_model._use_mixing_reg = True if self.gen_model.pct_mixing_reg else False

    # Create `StyleDiscriminator` type object by converting `ProDiscriminator` class into `StyleDiscriminator` class:
    StyleDiscriminator = type( 'StyleDiscriminator', ( StyleGAN, ), dict( ProDiscriminator.__dict__ ) )
    self.disc_model_metadata = checkpoint[ 'disc_model_metadata' ]
    self.disc_model_downsampler = self.disc_model_metadata[ 'disc_model_downsampler' ]
    self.num_classes_disc = self.disc_model_metadata[ 'num_classes_disc' ]
    self.disc_model = StyleDiscriminator(
      final_res = self.config.res_samples,
      pooler = self.disc_model_downsampler,
      blur_type = self.config.blur_type,
      nl = self.nl,
      num_classes = self.num_classes_disc,
      equalized_lr = self.config.use_equalized_lr,
      mbstd_group_size = self.config.mbstd_group_size
    )

    assert self.config.init_res <= self.config.res_samples

    # If pretrained model started at a higher resolution than 4:
    _curr_res = checkpoint[ 'curr_res' ]
    if _curr_res > 4:
      _init_res_log2 = int( np.log2( _curr_res ) )
      if float( _curr_res ) != 2**_init_res_log2:
        raise ValueError( 'Only resolutions that are powers of 2 are supported.' )
      num_scale_inc = _init_res_log2 - 2
      for _ in range( num_scale_inc ):
        self.gen_model.increase_scale()
        self.disc_model.increase_scale()

    # this applies to both networks simultaneously and takes care of fade_in_phase
    self.gen_model.alpha = checkpoint[ 'alpha' ]

    # Generator and Discriminator state data must match:
    assert self.gen_model.cls_base.__dict__ == \
           self.disc_model.cls_base.__dict__

    # Initialize EWMA Generator Model:
    self.gen_model_lagged = None
    if self.config.use_ewma_gen:
      _orig_mode = self.gen_model.training
      self.gen_model.train()
      self.gen_model.to( 'cpu' )
      with torch.no_grad():
        self.gen_model_lagged = copy.deepcopy( self.gen_model )  # for memory efficiency in GPU
        self.gen_model_lagged.to( self.config.metrics_dev )
        self.gen_model_lagged.train( mode = _orig_mode )
      self.gen_model.train( mode = _orig_mode )

    self.gen_model.to( self.config.dev )
    self.gen_model.load_state_dict( checkpoint[ 'gen_model_state_dict' ] )
    self.gen_model.zero_grad()
    if self.config.use_ewma_gen:
      self.gen_model_lagged.load_state_dict( checkpoint[ 'gen_model_lagged_state_dict' ] )
      self.gen_model.w_ewma = checkpoint[ 'w_ewma_lagged' ]
      self.gen_model_lagged.to( self.config.metrics_dev )
      self.gen_model_lagged.zero_grad()
    self.disc_model.to( self.config.dev )
    self.disc_model.load_state_dict( checkpoint[ 'disc_model_state_dict' ] )
    self.disc_model.zero_grad()

    self.sched_stop_step = checkpoint[ 'sched_stop_step' ]
    self.lr_sched = checkpoint[ 'lr_sched' ]
    self._scheduler_gen_state_dict = checkpoint['scheduler_gen_state_dict']
    self._scheduler_disc_state_dict = checkpoint['scheduler_disc_state_dict']
    self._sched_state_dict_set = False

    self.optimizer = checkpoint['optimizer']
    self.opt_gen.load_state_dict( checkpoint['opt_gen_state_dict'] )
    self.opt_disc.load_state_dict( checkpoint['opt_disc_state_dict'] )

    self.batch_size = checkpoint['batch_size']

    self.loss = checkpoint['loss']

    self.gradient_penalty = checkpoint['gradient_penalty']

    self.curr_dataset_batch_num = checkpoint['curr_dataset_batch_num']
    self.curr_epoch_num = checkpoint['curr_epoch_num']
    self.tot_num_epochs = checkpoint['tot_num_epochs']
    self.dataset_sz = checkpoint['dataset_sz']

    self.ac = checkpoint['ac']
    self.cond_gen = checkpoint['cond_gen']
    self._tensor = torch.FloatTensor
    if self.config.dev == torch.device( 'cuda' ):
      self._tensor = torch.cuda.FloatTensor
    if self.cond_gen:
      self.labels_one_hot_disc = self._tensor( self.batch_size, self.num_classes )
      self.labels_one_hot_gen = self._tensor( self.batch_size * self.config.gen_bs_mult, self.num_classes )
    self.cond_disc = checkpoint['cond_disc']

    self.valid_z = checkpoint['valid_z'].to( self.config.metrics_dev )
    self.valid_label = checkpoint['valid_label']
    if self.valid_label is not None:
      self.valid_label = self.valid_label.to( 'cpu' )
    self.grid_inputs_constructed = checkpoint['grid_inputs_constructed']
    self.rand_idxs = checkpoint['rand_idxs']
    self.gen_metrics_num = checkpoint['gen_metrics_num']
    self.disc_metrics_num = checkpoint['disc_metrics_num']

    self.curr_img_num = checkpoint[ 'curr_img_num' ]
    self.nimg_transition_lst = checkpoint[ 'nimg_transition_lst' ]

    self.not_trained_yet = checkpoint['not_trained_yet']

    # By default, use the pretrained model's statistics (you can change this by changing ds_mean and ds_std manually)
    self.ds_mean = checkpoint['ds_mean']
    self.ds_std = checkpoint['ds_std']
    if self._is_data_configed:
      self.data_config.ds_mean = self.ds_mean.squeeze().tolist()
      self.data_config.ds_std = self.ds_std.squeeze().tolist()

    self.latent_distribution = checkpoint[ 'latent_distribution' ]
    self.curr_phase_num = checkpoint[ 'curr_phase_num' ]
    self.eps = False
    if self.config.eps_drift > 0:
      self.eps = True
    self.lagged_params = checkpoint[ 'lagged_params' ]
    self._progressively_grow = checkpoint[ 'progressively_grow' ]


    self.pretrained_model = True

    # Print configuration:
    print( '---------- Loaded Model Configuration -----------' )
    print( self.config )
    print( '-------------------------------------------------' )
    print( "\n  If you would like to change any of the above configurations,\n" + \
           "  please do so via setting the attributes of your instantiated StyleGANLearner().config object.\n" )


# -*- coding: UTF-8 -*-

import os.path
from setuptools import setup

def readme( ):

  with open( os.path.abspath(
    os.path.join(
      os.path.dirname( __file__ ),
      'README.md' ) ) ) as f:

    return f.read( )

setup(
  name = 'gan-lab',
  version = '0.4.2',
  author = 'Sidhartha Parhi',
  author_email = 'sidhartha.parhi@gmail.com',
  description = "StyleGAN, ProGAN, and ResNet GANs to experiment with",
  long_description = readme( ),
  long_description_content_type = 'text/markdown',
  keywords = 'GAN GAN-Zoo ML generative neural model',
  url = "https://github.com/sidward14/gan-lab",
  packages = [
    'gan_lab',
    'gan_lab.utils',
    'gan_lab.resnetgan',
    'gan_lab.progan',
    'gan_lab.stylegan',
  ],
  dependency_links = [ ],
  install_requires = [
    'numpy >= 1.17.2',
    'scipy',
    'pillow >= 6.2.0',
    'matplotlib >= 3.1.1',
    'indexed >= 1.1.0',
    'torch >= 1.3.0',
    'torchvision >= 0.4.1',
    'lmdb >= 0.97',
    'tqdm',
  ],
  python_requires = '>= 3.6',
  include_package_data = True,
  zip_safe = False
)



# -*- coding: UTF-8 -*-

"""Base architecture class for ProGANs (Progressively Growing GANs).
"""

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

from abc import ABC, abstractmethod

import numpy as np
import torch
from torch import nn

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

FMAP_BASE = 8192
FMAP_MAX = 512

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++#

class ProGAN( nn.Module, ABC ):
  """This base class updates and keeps track of metadata that define the current
     state of the ProGAN model.
  """

  # these class attributes should be updated at the ProGANLearner class level

  _alpha = 1  # fading-in parameter
  _alpha_tol = 1.e-8

  _final_res = None
  _prev_res = None
  _curr_res = 4
  _scale_stage = int( np.log2( _curr_res ) ) - 1
  _fmap_prev = None
  _fmap = min( int( FMAP_BASE / ( 2**_scale_stage ) ), FMAP_MAX )
  _scale_inc_metadata_updated = False
  _fade_in_phase = False

  @classmethod
  def reset_state( cls ):
    """Call this to reset ProGAN state in case one wants to start over."""
    cls._alpha = 1
    cls._alpha_tol = 1.e-8

    cls._prev_res = None
    cls._curr_res = 4
    if cls._final_res is not None: assert cls._curr_res <= cls._final_res
    cls._scale_stage = int( np.log2( cls._curr_res ) ) - 1
    cls._fmap_prev = None
    cls._fmap = min( int( FMAP_BASE / ( 2**cls._scale_stage ) ), FMAP_MAX )
    cls._scale_inc_metadata_updated = False
    cls._fade_in_phase = False

  def __init__( self, final_res ):
    self.cls_base = self.__class__.__base__
    super( self.cls_base, self ).__init__()

    self.final_res = final_res
    assert self.curr_res <= self.final_res

  def increase_scale( self ):
    """Use this to increase scale during training or for initial resolution."""
    self.prev_res = self.curr_res
    self.curr_res = int( 2**( int( np.log2( self.curr_res ) ) + 1 ) )

    self.scale_stage = int( np.log2( self.curr_res ) ) - 1  # done this way for readability and congruency

    self.fmap_prev = self.fmap
    self.fmap = self.get_fmap( scale_stage = self.scale_stage )

    self.scale_inc_metadata_updated = True

    self.fade_in_phase = True

  def get_fmap( self, scale_stage ):
    return min( int( FMAP_BASE / ( 2**scale_stage ) ), FMAP_MAX )

  def most_parameters( self, recurse = True, excluded_params:list = [] ):
    """`torch.nn.Module.parameters()` generator method but with the option to exclude specified parameters."""
    for name, params in self.named_parameters( recurse = recurse ):
      if name not in excluded_params:
        yield params

  @property
  def fade_in_phase( self ):
    return self.cls_base._fade_in_phase

  @fade_in_phase.setter
  def fade_in_phase( self, new_fade_in_phase ):
    self.cls_base._fade_in_phase = new_fade_in_phase

  @property
  def scale_inc_metadata_updated( self ):
    return self.cls_base._scale_inc_metadata_updated

  @scale_inc_metadata_updated.setter
  def scale_inc_metadata_updated( self, new_scale_inc_metadata_updated ):
    self.cls_base._scale_inc_metadata_updated = new_scale_inc_metadata_updated

  @property
  def fmap( self ):
    return self.cls_base._fmap

  @fmap.setter
  def fmap( self, new_fmap ):
    self.cls_base._fmap = new_fmap

  @property
  def fmap_prev( self ):
    return self.cls_base._fmap_prev

  @fmap_prev.setter
  def fmap_prev( self, new_fmap_prev ):
    self.cls_base._fmap_prev = new_fmap_prev

  @property
  def scale_stage( self ):
    return self.cls_base._scale_stage

  @scale_stage.setter
  def scale_stage( self, new_scale_stage ):
    self.cls_base._scale_stage = new_scale_stage

  @property
  def curr_res( self ):
    return self.cls_base._curr_res

  @curr_res.setter
  def curr_res( self, new_curr_res ):
    self.cls_base._curr_res = new_curr_res

  @property
  def final_res( self ):
    return self.cls_base._final_res

  @final_res.setter
  def final_res( self, new_final_res ):
    self.cls_base._final_res = new_final_res

  @property
  def prev_res( self ):
    return self.cls_base._prev_res

  @prev_res.setter
  def prev_res( self, new_prev_res ):
    self.cls_base._prev_res = new_prev_res

  @property
  def alpha_tol( self ):
    return self.cls_base._alpha_tol

  @alpha_tol.setter
  def alpha_tol( self, new_alpha_tol ):
    self.cls_base._alpha_tol = new_alpha_tol

  @property
  def alpha( self ):
    return self.cls_base._alpha

  @alpha.setter
  def alpha( self, new_alpha ):
    if not ( 0. <= new_alpha < 1. + self.alpha_tol ):
      raise ValueError( 'Input alpha parameter must be in the range [0,1].' )

    if 1. - self.alpha_tol < new_alpha < 1. + self.alpha_tol:
      self.fade_in_phase = False
      self.cls_base._alpha = 1
    else:
      self.cls_base._alpha = new_alpha

  @abstractmethod
  def forward( self, x ):
    raise NotImplementedError( 'Can only call `forward` on valid subclasses.' )